'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.normalizeNodeWith = normalizeNodeWith;
exports.normalizeParentsWith = normalizeParentsWith;
exports.normalizeWith = normalizeWith;
exports.normalize = normalize;
exports.normalizeDocument = normalizeDocument;
exports.normalizeNodeByKey = normalizeNodeByKey;
exports.normalizeParentsByKey = normalizeParentsByKey;
exports.normalizeSelection = normalizeSelection;

var _warning = require('../utils/warning');

var _warning2 = _interopRequireDefault(_warning);

var _schema = require('../plugins/schema');

var _schema2 = _interopRequireDefault(_schema);

var _normalize = require('../utils/normalize');

var _normalize2 = _interopRequireDefault(_normalize);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// Maximum recursive calls for normalization
var MAX_CALLS = 50;

/**
 * Normalize a node (itself and its children) using a schema.
 *
 * @param  {Transform} transform
 * @param  {Schema} schema
 * @param  {Node} node
 * @return {Transform}
 */

function normalizeNodeWith(transform, schema, node) {
  // For performance considerations, we will check if the transform was changed
  var opCount = transform.operations.length;

  // Iterate over its children
  normalizeChildrenWith(transform, schema, node);

  var hasChanged = transform.operations.length != opCount;
  if (hasChanged) {
    // Refresh the node reference
    node = refreshNode(transform, node);
  }

  // Now normalize the node itself if it still exist
  if (node) {
    normalizeNodeOnly(transform, schema, node);
  }

  return transform;
}

/**
 * Normalize a node its parents using a schema.
 *
 * @param  {Transform} transform
 * @param  {Schema} schema
 * @param  {Node} node
 * @return {Transform}
 */

function normalizeParentsWith(transform, schema, node) {
  normalizeNodeOnly(transform, schema, node);

  // Normalize went back up to the document
  if (node.kind == 'document') {
    return transform;
  }

  // We search for the new parent
  node = refreshNode(transform, node);
  if (!node) {
    return transform;
  }

  var state = transform.state;
  var document = state.document;

  var parent = document.getParent(node.key);

  return normalizeParentsWith(transform, schema, parent);
}

/**
 * Normalize state using a schema.
 *
 * @param  {Transform} transform
 * @param  {Schema} schema
 * @return {Transform} transform
 */

function normalizeWith(transform, schema) {
  var state = transform.state;
  var document = state.document;


  if (!schema.hasValidators) {
    // Schema has no normalization rules
    return transform;
  }

  return transform.normalizeNodeWith(schema, document);
}

/**
 * Normalize the state using the core schema.
 *
 * @param  {Transform} transform
 * @return {Transform} transform
 */

function normalize(transform) {
  transform.normalizeDocument().normalizeSelection();
  return transform;
}

/**
 * Normalize only the document
 *
 * @param  {Transform} transform
 * @return {Transform} transform
 */

function normalizeDocument(transform) {
  return transform.normalizeWith(_schema2.default);
}

/**
 * Normalize a node and its children using core schema
 *
 * @param  {Transform} transform
 * @param  {Node or String} key
 * @return {Transform} transform
 */

function normalizeNodeByKey(transform, key) {
  key = _normalize2.default.key(key);
  var state = transform.state;
  var document = state.document;


  var node = document.key == key ? document : document.assertDescendant(key);

  transform.normalizeNodeWith(_schema2.default, node);
  return transform;
}

/**
 * Normalize a node and its parent using core schema
 *
 * @param  {Transform} transform
 * @param  {Node or String} key
 * @return {Transform} transform
 */

function normalizeParentsByKey(transform, key) {
  key = _normalize2.default.key(key);
  var state = transform.state;
  var document = state.document;

  var node = document.key == key ? document : document.assertDescendant(key);

  transform.normalizeParentsWith(_schema2.default, node);
  return transform;
}

/**
 * Normalize only the selection.
 *
 * @param  {Transform} transform
 * @return {Transform} transform
 */

function normalizeSelection(transform) {
  var state = transform.state;
  var _state = state,
      document = _state.document,
      selection = _state.selection;

  selection = selection.normalize(document);

  // If the selection is nulled (not normal)
  if (selection.isUnset || !document.hasDescendant(selection.anchorKey) || !document.hasDescendant(selection.focusKey)) {
    (0, _warning2.default)('Selection was invalid and reset to start of the document');
    var firstText = document.getFirstText();
    selection = selection.merge({
      anchorKey: firstText.key,
      anchorOffset: 0,
      focusKey: firstText.key,
      focusOffset: 0,
      isBackward: false
    });
  }

  state = state.merge({ selection: selection });
  transform.state = state;
  return transform;
}

/**
 * Refresh a reference to a node that have been modified in a transform.
 * @param  {Transform} transform
 * @param  {Node} node
 * @return {Node} newNode
 */

function refreshNode(transform, node) {
  var state = transform.state;
  var document = state.document;


  if (node.kind == 'document') {
    return document;
  }

  return document.getDescendant(node.key);
}

/**
 * Normalize all children of a node
 * @param  {Transform} transform
 * @param  {Schema} schema
 * @param  {Node} node
 * @return {Transform} transform
 */

function normalizeChildrenWith(transform, schema, node) {
  if (node.kind == 'text') {
    return transform;
  }

  return node.nodes.reduce(function (t, child) {
    return t.normalizeNodeWith(schema, child);
  }, transform);
}

/**
 * Normalize a node, but not its children
 *
 * @param  {Transform} transform
 * @param  {Schema} schema
 * @param  {Node} node
 * @return {Transform} transform
 */

function normalizeNodeOnly(transform, schema, node) {
  var recursiveCount = 0;

  // Auxiliary function, called recursively, with a maximum calls safety net.
  function _recur(_transform, _node) {
    // _node.validate should be memoized
    var failure = _node.validate(schema);

    // Node is valid?
    if (!failure) {
      return _transform;
    }

    var value = failure.value,
        rule = failure.rule;

    // Normalize and get the new state

    rule.normalize(_transform, _node, value);

    // Search for the updated node in the new state
    var newNode = refreshNode(_transform, _node);

    // Node no longer exist, go back to normalize parents
    if (!newNode) {
      return _transform;
    }

    recursiveCount++;
    if (recursiveCount > MAX_CALLS) {
      throw new Error('Unexpected number of successive normalizations. Aborting.');
    }

    return _recur(_transform, newNode);
  }

  return _recur(transform, node);
}